@@ -3,9 +3,10 @@ import logging
 import warnings
 from collections import defaultdict
 from contextlib import suppress
+from inspect import isawaitable
 from ipaddress import ip_address
 from time import time
-from typing import Callable, Dict, Optional, Type, TypeVar
+from typing import Callable, Dict, Generator, Optional, Tuple, Type, TypeVar
 
 from playwright.async_api import (
     BrowserContext,
@@ -105,8 +106,12 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             self.context_kwargs[name].update(kwargs)
         if "default" not in self.context_kwargs and default_context_kwargs:
             self.context_kwargs["default"] = default_context_kwargs
-        self.contexts: Dict[str, BrowserContext] = dict()
-        self.context_semaphores: Dict[str, asyncio.Semaphore] = dict()
+        self.contexts: Dict[str, BrowserContext] = {}
+        self.context_semaphores: Dict[str, asyncio.Semaphore] = {}
+
+        self.abort_request: Optional[Callable[[PlaywrightRequest], bool]] = None
+        if crawler.settings.get("PLAYWRIGHT_ABORT_REQUEST"):
+            self.abort_request = load_object(crawler.settings["PLAYWRIGHT_ABORT_REQUEST"])
 
     @classmethod
     def from_crawler(cls: Type[PlaywrightHandler], crawler: Crawler) -> PlaywrightHandler:
@@ -131,8 +136,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         )
         self.contexts = dict(zip(self.context_kwargs.keys(), contexts))
         self.context_semaphores = {
-            name: asyncio.Semaphore(value=self.max_pages_per_context)
-            for name in self.contexts.keys()
+            name: asyncio.Semaphore(value=self.max_pages_per_context) for name in self.contexts
         }
 
     async def _create_browser_context(self, name: str, context_kwargs: dict) -> BrowserContext:
@@ -251,9 +255,14 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             page_coroutines = page_coroutines.values()
         for pc in page_coroutines:
             if isinstance(pc, PageCoroutine):
-                method = getattr(page, pc.method)
-                pc.result = await method(*pc.args, **pc.kwargs)
-                await page.wait_for_load_state(timeout=self.default_navigation_timeout)
+                try:
+                    method = getattr(page, pc.method)
+                except AttributeError:
+                    logger.warning(f"Ignoring {repr(pc)}: could not find coroutine")
+                else:
+                    result = method(*pc.args, **pc.kwargs)
+                    pc.result = await result if isawaitable(result) else result
+                    await page.wait_for_load_state(timeout=self.default_navigation_timeout)
             else:
                 logger.warning(
                     f"Ignoring {repr(pc)}: expected PageCoroutine, got {repr(type(pc))}"
@@ -278,8 +287,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
 
         headers = Headers(response.headers)
         headers.pop("Content-Encoding", None)
-        encoding = _get_response_encoding(headers, body_str) or "utf-8"
-        body = body_str.encode(encoding)
+        body, encoding = _encode_body(headers=headers, text=body_str)
         respcls = responsetypes.from_args(headers=headers, url=page.url, body=body)
         return respcls(
             url=page.url,
@@ -292,13 +300,6 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             ip_address=server_ip_address,
         )
 
-    def _make_close_page_callback(self, context_name: str) -> Callable:
-        def close_page_callback() -> None:
-            if context_name in self.context_semaphores:
-                self.context_semaphores[context_name].release()
-
-        return close_page_callback
-
     def _increment_request_stats(self, request: PlaywrightRequest) -> None:
         stats_prefix = "playwright/request_count"
         self.stats.inc_value(stats_prefix)
@@ -313,6 +314,13 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         self.stats.inc_value(f"{stats_prefix}/resource_type/{response.request.resource_type}")
         self.stats.inc_value(f"{stats_prefix}/method/{response.request.method}")
 
+    def _make_close_page_callback(self, context_name: str) -> Callable:
+        def close_page_callback() -> None:
+            if context_name in self.context_semaphores:
+                self.context_semaphores[context_name].release()
+
+        return close_page_callback
+
     def _make_close_browser_context_callback(self, name: str) -> Callable:
         def close_browser_context_callback() -> None:
             logger.debug("Browser context closed: '%s'", name)
@@ -328,6 +336,11 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     ) -> Callable:
         async def _request_handler(route: Route, playwright_request: PlaywrightRequest) -> None:
             """Override request headers, method and body."""
+            if self.abort_request and self.abort_request(playwright_request):
+                await route.abort()
+                self.stats.inc_value("playwright/request_count/aborted")
+                return None
+
             processed_headers = await self.process_request_headers(
                 self.browser_type, playwright_request, scrapy_headers
             )
@@ -347,11 +360,19 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         return _request_handler
 
 
-def _get_response_encoding(headers: Headers, body: str) -> Optional[str]:
-    encoding = None
+def _possible_encodings(headers: Headers, text: str) -> Generator[str, None, None]:
     if headers.get("content-type"):
         content_type = to_unicode(headers["content-type"])
-        encoding = http_content_type_encoding(content_type)
-    if not encoding:
-        encoding = html_body_declared_encoding(body)
-    return encoding
+        yield http_content_type_encoding(content_type)
+    yield html_body_declared_encoding(text)
+
+
+def _encode_body(headers: Headers, text: str) -> Tuple[bytes, str]:
+    for encoding in filter(None, _possible_encodings(headers, text)):
+        try:
+            body = text.encode(encoding)
+        except UnicodeEncodeError:
+            pass
+        else:
+            return body, encoding
+    return text.encode("utf-8"), "utf-8"  # fallback
